/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file aac.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 24.02.2006
 *
 * Functinality for controlling aac placyback using faad and ao. With the aac_play
 * function we can specify a mp4/aac file and start and stop offsets. The audio_play
 * function starts decoding the aac file.
 * Much of this file are copyed / inspired by the (minimalistic) FAAD API
 * and the FAAD frontend code.
 * 
 * To build with AAC support, this is required:
 * - faad2-0 and faad2-dev installed.
 * - faad and mp4ff must be linked during build. "gcc (....) -lfaad -lmp4ff"
 * 
 * TODO: Clean up! There's too much code here wich we don't use.
 * NOTE: This AAC decoder are not yet fully implemented, so AAC support in libdaisy are not yet complete.
 */
#ifdef WITH_AAC				/** check if we want to compile with AAC support */

#include <faad.h>
#include <string.h>
#include <mp4ff.h>			/** to process the mp4 file format */
#include <stdbool.h>		/** bool */

#include "aac.h"			/** daisy aac decoder interface */
#include "audio.h"			/** daisy audio module */
#include "faad2/audio.h"	/** FAAD2 AAC decoder. This is borrowed from the FAAD2 fronted. */
#include "report.h"			/** reporter */
#include "common.h"			/** Common project settings */



/* ************************************************	*
 * 	Global type definitions and datastructures		*
 * ************************************************	*/

#define AAC_MAX_CHANNELS 6 /* make this higher to support files with
	                          more channels */

#ifndef min
#define min(a,b) ( (a) < (b) ? (a) : (b) )
#endif



/* FAAD file buffering routines */
typedef struct {
    long bytes_into_buffer;
    long bytes_consumed;
    long file_offset;
    unsigned char *buffer;
    int at_eof;
    FILE *infile;
} aac_buffer;


/* MicroSoft channel definitions */
#define SPEAKER_FRONT_LEFT				0x1
#define SPEAKER_FRONT_RIGHT				0x2
#define SPEAKER_FRONT_CENTER			0x4
#define SPEAKER_LOW_FREQUENCY			0x8
#define SPEAKER_BACK_LEFT				0x10
#define SPEAKER_BACK_RIGHT				0x20
#define SPEAKER_FRONT_LEFT_OF_CENTER	0x40
#define SPEAKER_FRONT_RIGHT_OF_CENTER	0x80
#define SPEAKER_BACK_CENTER				0x100
#define SPEAKER_SIDE_LEFT				0x200
#define SPEAKER_SIDE_RIGHT				0x400
#define SPEAKER_TOP_CENTER				0x800
#define SPEAKER_TOP_FRONT_LEFT			0x1000
#define SPEAKER_TOP_FRONT_CENTER		0x2000
#define SPEAKER_TOP_FRONT_RIGHT			0x4000
#define SPEAKER_TOP_BACK_LEFT			0x8000
#define SPEAKER_TOP_BACK_CENTER			0x10000
#define SPEAKER_TOP_BACK_RIGHT			0x20000
#define SPEAKER_RESERVED				0x80000000

static int adts_sample_rates[] = {96000,88200,64000,48000,44100,32000,24000,22050,16000,12000,11025,8000,7350,0,0,0};



/* ************************************************	*
 * 			Static function declarations 			*
 * ************************************************	*/
 
static
int GetAACTrack(mp4ff_t *infile);

static
int decodeMP4file(struct_audio_data_t *data, char *mp4file, char *sndfile, char *adts_fn, int to_stdout,
                  int outputFormat, int fileType, int downMatrix, int noGapless,
                  int infoOnly, int adts_out, float *song_length);

static
int decodeAACfile(struct_audio_data_t *data, char *aacfile, char *sndfile, char *adts_fn, int to_stdout,
                  int def_srate, int object_type, int outputFormat, int fileType,
                  int downMatrix, int infoOnly, int adts_out, int old_format,
                  float *song_length);

static
uint32_t read_callback(void *user_data, void *buffer, uint32_t length);

static
uint32_t seek_callback(void *user_data, uint64_t position);

static
int adts_parse(aac_buffer *b, int *bitrate, float *length);

static
void advance_buffer(aac_buffer *b, int bytes);

static
int fill_buffer(aac_buffer *b);

static
unsigned char *MakeAdtsHeader(int *dataSize, faacDecFrameInfo *hInfo, int old_format);

static
int FindAdtsSRIndex(int sr);

static
long aacChannelConfig2wavexChannelMask(faacDecFrameInfo *hInfo);

static
char *position2string(int position);

static
int GetAACTrack(mp4ff_t *infile);



/* ************************************************	*
 * 			Static function definitions 			*
 * ************************************************	*/

/*
 * Finds out wich track in an MP4 file that is the AAC track.
 * @parm mp4ff_t *infile - the  MP4 file.
 */
static
int GetAACTrack(mp4ff_t *infile)
{
    /* find AAC track */
    int i, rc;
    int numTracks = mp4ff_total_tracks(infile);

    for (i = 0; i < numTracks; i++)
    {
        unsigned char *buff = NULL;
        unsigned int buff_size = 0;
        mp4AudioSpecificConfig mp4ASC;

        mp4ff_get_decoder_config(infile, i, &buff, &buff_size);

        if (buff)
        {
            rc = AudioSpecificConfig(buff, buff_size, &mp4ASC);
            free(buff);

            if (rc < 0)
                continue;
            return i;
        }
    }

    /* can't decode this */
    return -1;
}

static
long aacChannelConfig2wavexChannelMask(faacDecFrameInfo *hInfo)
{
    if (hInfo->channels == 6 && hInfo->num_lfe_channels)
    {
        return SPEAKER_FRONT_LEFT + SPEAKER_FRONT_RIGHT +
            SPEAKER_FRONT_CENTER + SPEAKER_LOW_FREQUENCY +
            SPEAKER_BACK_LEFT + SPEAKER_BACK_RIGHT;
    } else {
        return 0;
    }
}

/**
 * Finds the samperate's index in the adts_sample_rates array.
 * @param int sr - the samplerate to look up.
 * @return int - the samplerates index in the array. 15 if no match were found.
 */
static
int FindAdtsSRIndex(int sr)
{
    int i;

    for (i = 0; i < 16; i++)
    {
        if (sr == adts_sample_rates[i])
            return i;
    }
    return 16 - 1;
}

/**
 * Creates an Adts header.
 * @param int *dataSize - size of the data part.
 * @param faacDecFrameInfo *hInfo - info about the current frame.
 * @param int old_format - what format we will use.
 * TODO: We probably don't need this in libdaisy to simple decode MP4/AAC audio. All code related to ADTS can probably be removed.
 */
static
unsigned char *MakeAdtsHeader(int *dataSize, faacDecFrameInfo *hInfo, int old_format)
{
    unsigned char *data;
    int profile = (hInfo->object_type - 1) & 0x3;
    int sr_index = ((hInfo->sbr == SBR_UPSAMPLED) || (hInfo->sbr == NO_SBR_UPSAMPLED)) ?
        FindAdtsSRIndex(hInfo->samplerate / 2) : FindAdtsSRIndex(hInfo->samplerate);
    int skip = (old_format) ? 8 : 7;
    int framesize = skip + hInfo->bytesconsumed;

    if (hInfo->header_type == ADTS)
        framesize -= skip;

    *dataSize = 7;

    data = malloc(*dataSize * sizeof(unsigned char));
    memset(data, 0, *dataSize * sizeof(unsigned char));

    data[0] += 0xFF; /* 8b: syncword */

    data[1] += 0xF0; /* 4b: syncword */
    /* 1b: mpeg id = 0 */
    /* 2b: layer = 0 */
    data[1] += 1; /* 1b: protection absent */

    data[2] += ((profile << 6) & 0xC0); /* 2b: profile */
    data[2] += ((sr_index << 2) & 0x3C); /* 4b: sampling_frequency_index */
    /* 1b: private = 0 */
    data[2] += ((hInfo->channels >> 2) & 0x1); /* 1b: channel_configuration */

    data[3] += ((hInfo->channels << 6) & 0xC0); /* 2b: channel_configuration */
    /* 1b: original */
    /* 1b: home */
    /* 1b: copyright_id */
    /* 1b: copyright_id_start */
    data[3] += ((framesize >> 11) & 0x3); /* 2b: aac_frame_length */

    data[4] += ((framesize >> 3) & 0xFF); /* 8b: aac_frame_length */

    data[5] += ((framesize << 5) & 0xE0); /* 3b: aac_frame_length */
    data[5] += ((0x7FF >> 6) & 0x1F); /* 5b: adts_buffer_fullness */

    data[6] += ((0x7FF << 2) & 0x3F); /* 6b: adts_buffer_fullness */
    /* 2b: num_raw_data_blocks */

    return data;
}

/**
 * Fills up a buffer.
 * @param aac_buffer *b - pointer to a buffer.
 * @return int - returns 1
 */
static
int fill_buffer(aac_buffer *b)
{
    int bread;

    if (b->bytes_consumed > 0)
    {
        if (b->bytes_into_buffer)
        {
            memmove((void*)b->buffer, (void*)(b->buffer + b->bytes_consumed),
                b->bytes_into_buffer*sizeof(unsigned char));
        }

        if (!b->at_eof)
        {
            bread = fread((void*)(b->buffer + b->bytes_into_buffer), 1,
                b->bytes_consumed, b->infile);

            if (bread != b->bytes_consumed)
                b->at_eof = 1;

            b->bytes_into_buffer += bread;
        }

        b->bytes_consumed = 0;

        if (b->bytes_into_buffer > 3)
        {
            if (memcmp(b->buffer, "TAG", 3) == 0)
                b->bytes_into_buffer = 0;
        }
        if (b->bytes_into_buffer > 11)
        {
            if (memcmp(b->buffer, "LYRICSBEGIN", 11) == 0)
                b->bytes_into_buffer = 0;
        }
        if (b->bytes_into_buffer > 8)
        {
            if (memcmp(b->buffer, "APETAGEX", 8) == 0)
                b->bytes_into_buffer = 0;
        }
    }

    return 1;
}

/**
 * Advances a number of bytes in a buffer.
 * @param aac_buffer *b - Pointer to a buffer.
 * @param int bytes - Number of bytes to advance.
 */
static
void advance_buffer(aac_buffer *b, int bytes)
{
    b->file_offset += bytes;
    b->bytes_consumed = bytes;
    b->bytes_into_buffer -= bytes;
}

/**
 * Parses adts data.
 * @param aac_buffer *b - pointer to buffer.
 * @param int *bitrate - audio bitrate
 * @param float *length - audio length
 * @return int - returns 1
 */
static
int adts_parse(aac_buffer *b, int *bitrate, float *length)
{
    int frames, frame_length;
    int t_framelength = 0;
    int samplerate;
    float frames_per_sec, bytes_per_frame;

    /* Read all frames to ensure correct time and bitrate */
    for (frames = 0; /* */; frames++)
    {
        fill_buffer(b);

        if (b->bytes_into_buffer > 7)
        {
            /* check syncword */
            if (!((b->buffer[0] == 0xFF)&&((b->buffer[1] & 0xF6) == 0xF0)))
                break;

            if (frames == 0)
                samplerate = adts_sample_rates[(b->buffer[2]&0x3c)>>2];

            frame_length = ((((unsigned int)b->buffer[3] & 0x3)) << 11)
                | (((unsigned int)b->buffer[4]) << 3) | (b->buffer[5] >> 5);

            t_framelength += frame_length;

            if (frame_length > b->bytes_into_buffer)
                break;

            advance_buffer(b, frame_length);
        } else {
            break;
        }
    }

    frames_per_sec = (float)samplerate/1024.0f;
    if (frames != 0)
        bytes_per_frame = (float)t_framelength/(float)(frames*1000);
    else
        bytes_per_frame = 0;
    *bitrate = (int)(8. * bytes_per_frame * frames_per_sec + 0.5);
    if (frames_per_sec != 0)
        *length = (float)frames/frames_per_sec;
    else
        *length = 1;

    return 1;
}

/**
 * Read callback function.
 * @param void *user_data - callback data pointer.
 * @param void *buffer - pointer to buffer
 * @param uint32_t length - length of data to read
 * @return uint32_t - returns number of items read.
 */
static
uint32_t read_callback(void *user_data, void *buffer, uint32_t length)
{
    return fread(buffer, 1, length, (FILE*)user_data);
}

/**
 * Seek callback function
 * @param void *user_data - callback data pointer.
 * @param uint64_t position - the position to seek to.
 * @return uint32_t - returns 0 on success, anything else on fail.
 */
static
uint32_t seek_callback(void *user_data, uint64_t position)
{
    return fseek((FILE*)user_data, position, SEEK_SET);
}

/**
 * Decodes an AAC file.
 * @param struct_audio_data_t *data - pointer to the audio data structure. Used in callback and ao output functions.
 * @param char *aacfile - pointer to aac file.
 * @param char *sndfile - pointer to sound file.
 * @param char *adts_fn - pointer to adts output filename. (TODO: Remove adts if we don't need it).
 * @param int to_stdout - output to stdout.
 * @param int def_srate - 
 * @param int object_type - 
 * @param int outputFormat - format of output data.
 * @param int fileType - 
 * @param int downMatrix - 
 * @param int infoOnly - only output info.
 * @param int adts_out - output as adts.
 * @param int old_format - use old adts format.
 * @param float *song_length - length of audio.
 */
static
int decodeAACfile(struct_audio_data_t *data, char *aacfile, char *sndfile, char *adts_fn, int to_stdout,
                  int def_srate, int object_type, int outputFormat, int fileType,
                  int downMatrix, int infoOnly, int adts_out, int old_format,
                  float *song_length)
{
    int tagsize;
    unsigned long samplerate;
    unsigned char channels;
    void *sample_buffer;

    audio_file *aufile;

    FILE *adtsFile;
    unsigned char *adtsData;
    int adtsDataSize;

    faacDecHandle hDecoder;
    faacDecFrameInfo frameInfo;
    faacDecConfigurationPtr config;

    char percents[200];
    int percent, old_percent = -1;
    int bread, fileread;
    int header_type = 0;
    int bitrate = 0;
    float length = 0;

    int first_time = 1;

    aac_buffer b;

    memset(&b, 0, sizeof(aac_buffer));

    if (adts_out)
    {
        adtsFile = fopen(adts_fn, "wb");
        if (adtsFile == NULL)
        {
            fprintf(stderr, "Error opening file: %s\n", adts_fn);
            return 1;
        }
    }

    b.infile = fopen(aacfile, "rb");
    if (b.infile == NULL)
    {
        /* unable to open file */
        fprintf(stderr, "Error opening file: %s\n", aacfile);
        return 1;
    }

    fseek(b.infile, 0, SEEK_END);
    fileread = ftell(b.infile);
    fseek(b.infile, 0, SEEK_SET);

    if (!(b.buffer = (unsigned char*)malloc(FAAD_MIN_STREAMSIZE*AAC_MAX_CHANNELS)))
    {
        fprintf(stderr, "Memory allocation error\n");
        return 0;
    }
    memset(b.buffer, 0, FAAD_MIN_STREAMSIZE*AAC_MAX_CHANNELS);

    bread = fread(b.buffer, 1, FAAD_MIN_STREAMSIZE*AAC_MAX_CHANNELS, b.infile);
    b.bytes_into_buffer = bread;
    b.bytes_consumed = 0;
    b.file_offset = 0;

    if (bread != FAAD_MIN_STREAMSIZE*AAC_MAX_CHANNELS)
        b.at_eof = 1;

    tagsize = 0;
    if (!memcmp(b.buffer, "ID3", 3))
    {
        /* high bit is not used */
        tagsize = (b.buffer[6] << 21) | (b.buffer[7] << 14) |
            (b.buffer[8] <<  7) | (b.buffer[9] <<  0);

        tagsize += 10;
        advance_buffer(&b, tagsize);
        fill_buffer(&b);
    }

    hDecoder = faacDecOpen();

    /* Set the default object type and samplerate */
    /* This is useful for RAW AAC files */
    config = faacDecGetCurrentConfiguration(hDecoder);
    if (def_srate)
        config->defSampleRate = def_srate;
    config->defObjectType = object_type;
    config->outputFormat = outputFormat;
    config->downMatrix = downMatrix;
    config->useOldADTSFormat = old_format;
    config->dontUpSampleImplicitSBR = 1;
    faacDecSetConfiguration(hDecoder, config);

    /* get AAC infos for printing */
    header_type = 0;
    if ((b.buffer[0] == 0xFF) && ((b.buffer[1] & 0xF6) == 0xF0))
    {
        adts_parse(&b, &bitrate, &length);
        fseek(b.infile, tagsize, SEEK_SET);

        bread = fread(b.buffer, 1, FAAD_MIN_STREAMSIZE*AAC_MAX_CHANNELS, b.infile);
        if (bread != FAAD_MIN_STREAMSIZE*AAC_MAX_CHANNELS)
            b.at_eof = 1;
        else
            b.at_eof = 0;
        b.bytes_into_buffer = bread;
        b.bytes_consumed = 0;
        b.file_offset = tagsize;

        header_type = 1;
    } else if (memcmp(b.buffer, "ADIF", 4) == 0) {
        int skip_size = (b.buffer[4] & 0x80) ? 9 : 0;
        bitrate = ((unsigned int)(b.buffer[4 + skip_size] & 0x0F)<<19) |
            ((unsigned int)b.buffer[5 + skip_size]<<11) |
            ((unsigned int)b.buffer[6 + skip_size]<<3) |
            ((unsigned int)b.buffer[7 + skip_size] & 0xE0);

        length = (float)fileread;
        if (length != 0)
        {
            length = ((float)length*8.f)/((float)bitrate) + 0.5f;
        }

        bitrate = (int)((float)bitrate/1000.0f + 0.5f);

        header_type = 2;
    }

    *song_length = length;

    fill_buffer(&b);
    if ((bread = faacDecInit(hDecoder, b.buffer,
        b.bytes_into_buffer, &samplerate, &channels)) < 0)
    {
        /* If some error initializing occured, skip the file */
        fprintf(stderr, "Error initializing decoder library.\n");
        if (b.buffer)
            free(b.buffer);
        faacDecClose(hDecoder);
        fclose(b.infile);
        return 1;
    }
    advance_buffer(&b, bread);
    fill_buffer(&b);

    /* print AAC file info */
    fprintf(stderr, "%s file info:\n", aacfile);
    switch (header_type)
    {
    case 0:
        fprintf(stderr, "RAW\n\n");
        break;
    case 1:
        fprintf(stderr, "ADTS, %.3f sec, %d kbps, %lu Hz\n\n",
            length, bitrate, samplerate);
        break;
    case 2:
        fprintf(stderr, "ADIF, %.3f sec, %d kbps, %lu Hz\n\n",
            length, bitrate, samplerate);
        break;
    }

    if (infoOnly)
    {
        faacDecClose(hDecoder);
        fclose(b.infile);
        if (b.buffer)
            free(b.buffer);
        return 0;
    }

    do
    {
        sample_buffer = faacDecDecode(hDecoder, &frameInfo,
            b.buffer, b.bytes_into_buffer);

        if (adts_out == 1)
        {
            int skip = (old_format) ? 8 : 7;
            adtsData = MakeAdtsHeader(&adtsDataSize, &frameInfo, old_format);

            /* write the adts header */
            if (DEBUG_AAC) report ("I write the adts header", REP_DEBUG);
            fwrite(adtsData, 1, adtsDataSize, adtsFile);

            /* write the frame data */
            if (DEBUG_AAC) report ("I write the adts frame data", REP_DEBUG);
            if (frameInfo.header_type == ADTS)
                fwrite(b.buffer + skip, 1, frameInfo.bytesconsumed - skip, adtsFile);
            else
                fwrite(b.buffer, 1, frameInfo.bytesconsumed, adtsFile);
        }

        /* update buffer indices */
        advance_buffer(&b, frameInfo.bytesconsumed);

        if (frameInfo.error > 0)
        {
            fprintf(stderr, "Error: %s\n",
                faacDecGetErrorMessage(frameInfo.error));
        }

        /* open the sound file now that the number of channels are known */
        if (first_time && !frameInfo.error)
        {

            if (!adts_out)
            {
                /* open output file */
                if (!to_stdout)
                {
                    aufile = open_audio_file(sndfile, frameInfo.samplerate, frameInfo.channels,
                        outputFormat, fileType, aacChannelConfig2wavexChannelMask(&frameInfo));
                } else {
                    aufile = open_audio_file("-", frameInfo.samplerate, frameInfo.channels,
                        outputFormat, fileType, aacChannelConfig2wavexChannelMask(&frameInfo));
                }
                if (aufile == NULL)
                {
                    if (b.buffer)
                        free(b.buffer);
                    faacDecClose(hDecoder);
                    fclose(b.infile);
                    return 0;
                }
            } else {
                fprintf(stderr, "Writing output MPEG-4 AAC ADTS file.\n\n");
            }
            first_time = 0;
        }

        if ((frameInfo.error == 0) && (frameInfo.samples > 0) && (!adts_out))
        {
        	if (DEBUG_AAC) report ("write_audio_file1", REP_DEBUG);
			write_audio_file(aufile, sample_buffer, frameInfo.samples, 0);
        }

        /* fill buffer */
        fill_buffer(&b);

        if (b.bytes_into_buffer == 0)
            sample_buffer = NULL; /* to make sure it stops now */

    } while (sample_buffer != NULL);

    faacDecClose(hDecoder);

    if (adts_out == 1)
    {
        fclose(adtsFile);
    }

    fclose(b.infile);

    if (!first_time && !adts_out)
        close_audio_file(aufile);

    if (b.buffer)
        free(b.buffer);

    return frameInfo.error;
}

/**
 * Decodes an MP4 file.
 * @param struct_audio_data_t *data - pointer to the audio data structure. Used in callback and ao output functions.
 * @param char *mp4file - pointer to the mp4-file.
 * @param char *sndfile - pointer to sound file.
 * @param char *adts_fn - pointer to adts output filename. (TODO: Remove adts if we don't need it).
 * @param int to_stdout - output to stdout.
 * @param int outputFormat - output format.
 * @param int fileType - 
 * @param int downMatrix - 
 * @param int noGapless - dont do gapless playback.
 * @param int infoOnly - only output file information.
 * @param int adts_out - output adts.
 * @param float *song_length - audio length.
 */
static
int decodeMP4file(struct_audio_data_t *data, char *mp4file, char *sndfile, char *adts_fn, int to_stdout,
                  int outputFormat, int fileType, int downMatrix, int noGapless,
                  int infoOnly, int adts_out, float *song_length)
{
    int track;
    unsigned long samplerate;
    unsigned char channels;
    void *sample_buffer;

    mp4ff_t *infile;
    long sampleId, numSamples;

    audio_file *aufile;

    FILE *mp4File;
    FILE *adtsFile;
    unsigned char *adtsData;
    int adtsDataSize;

    faacDecHandle hDecoder;
    faacDecConfigurationPtr config;
    faacDecFrameInfo frameInfo;
    mp4AudioSpecificConfig mp4ASC;

    unsigned char *buffer;
    int buffer_size;

    char percents[200];
    int percent, old_percent = -1;

    int first_time = 1;

    /* for gapless decoding */
    unsigned int useAacLength = 1;
    unsigned int initial = 1;
    unsigned int framesize;
    unsigned long timescale;


    /* initialise the callback structure */
    mp4ff_callback_t *mp4cb = malloc(sizeof(mp4ff_callback_t));

    mp4File = fopen(mp4file, "rb");
    mp4cb->read = read_callback;
    mp4cb->seek = seek_callback;
    mp4cb->user_data = mp4File;


    hDecoder = faacDecOpen();

    /* Set configuration */
    config = faacDecGetCurrentConfiguration(hDecoder);

    if (DEBUG_AAC) {
    	char msg[STRLEN];
	    snprintf (msg, STRLEN, "\noutputFormat: %d\ndownMatrix: %d", outputFormat, downMatrix);
	    report (msg, REP_DEBUG);
    }

    config->outputFormat = outputFormat;
    config->downMatrix = downMatrix;
    config->dontUpSampleImplicitSBR = 1;
    faacDecSetConfiguration(hDecoder, config);

    if (adts_out)
    {
        adtsFile = fopen(adts_fn, "wb");
        if (adtsFile == NULL)
        {
            fprintf(stderr, "Error opening file: %s\n", adts_fn);
            return 1;
        }
    }

    infile = mp4ff_open_read(mp4cb);
    if (!infile)
    {
        /* unable to open file */
        fprintf(stderr, "Error opening file: %s\n", mp4file);
        return 1;
    }

    if ((track = GetAACTrack(infile)) < 0)
    {
        fprintf(stderr, "Unable to find correct AAC sound track in the MP4 file.\n");
        faacDecClose(hDecoder);
        mp4ff_close(infile);
        free(mp4cb);
        fclose(mp4File);
        return 1;
    }

    buffer = NULL;
    buffer_size = 0;
    mp4ff_get_decoder_config(infile, track, &buffer, &buffer_size);

    if(faacDecInit2(hDecoder, buffer, buffer_size,
                    &samplerate, &channels) < 0)
    {
        /* If some error initializing occured, skip the file */
        fprintf(stderr, "Error initializing decoder library.\n");
        faacDecClose(hDecoder);
        mp4ff_close(infile);
        free(mp4cb);
        fclose(mp4File);
        return 1;
    }

	/**
	 * Initializing daisylibao now that we know samplerate and channels
	 */
	libao_initiate(AO_FMT_LITTLE, channels, samplerate);

    timescale = mp4ff_time_scale(infile, track);
    framesize = 1024;
    useAacLength = 0;

    if (buffer)
    {
        if (AudioSpecificConfig(buffer, buffer_size, &mp4ASC) >= 0)
        {
            if (mp4ASC.frameLengthFlag == 1) framesize = 960;
            if (mp4ASC.sbr_present_flag == 1) framesize *= 2;
        }
        free(buffer);
    }

    /* print some mp4 file info */
    fprintf(stderr, "%s file info:\n\n", mp4file);
    {
        char *tag = NULL, *item = NULL;
        int k, j;
        char *ot[6] = { "NULL", "MAIN AAC", "LC AAC", "SSR AAC", "LTP AAC", "HE AAC" };
        long samples = mp4ff_num_samples(infile, track);
        float f = 1024.0;
        float seconds;
        if (mp4ASC.sbr_present_flag == 1)
        {
            f = f * 2.0;
        }
        seconds = (float)samples*(float)(f-1.0)/(float)mp4ASC.samplingFrequency;

        *song_length = seconds;

        fprintf(stderr, "%s\t%.3f secs, %d ch, %d Hz\n\n", ot[(mp4ASC.objectTypeIndex > 5)?0:mp4ASC.objectTypeIndex],
            seconds, mp4ASC.channelsConfiguration, mp4ASC.samplingFrequency);

#define PRINT_MP4_METADATA
#ifdef PRINT_MP4_METADATA
        j = mp4ff_meta_get_num_items(infile);
        for (k = 0; k < j; k++)
        {
            if (mp4ff_meta_get_by_index(infile, k, &item, &tag))
            {
                if (item != NULL && tag != NULL)
                {
                    fprintf(stderr, "%s: %s\n", item, tag);
                    free(item); item = NULL;
                    free(tag); tag = NULL;
                }
            }
        }
        if (j > 0) fprintf(stderr, "\n");
#endif
    }

    if (infoOnly)
    {
        faacDecClose(hDecoder);
        mp4ff_close(infile);
        free(mp4cb);
        fclose(mp4File);
        return 0;
    }

    numSamples = mp4ff_num_samples(infile, track);

    for (sampleId = 0; sampleId < numSamples; sampleId++)
    {
        int rc;
        long dur;
        unsigned int sample_count;
        unsigned int delay = 0;

        /* get acces unit from MP4 file */
        buffer = NULL;
        buffer_size = 0;

        dur = mp4ff_get_sample_duration(infile, track, sampleId);

        rc = mp4ff_read_sample(infile, track, sampleId, &buffer,  &buffer_size);
        if (rc == 0)
        {
            fprintf(stderr, "Reading from MP4 file failed.\n");
            faacDecClose(hDecoder);
            mp4ff_close(infile);
            free(mp4cb);
            fclose(mp4File);
            return 1;
        }

        sample_buffer = faacDecDecode(hDecoder, &frameInfo, buffer, buffer_size);

        if (adts_out == 1)
        {
            adtsData = MakeAdtsHeader(&adtsDataSize, &frameInfo, 0);

            /* write the adts header */
            if (DEBUG_AAC) report ("I write the adts header", REP_DEBUG);
            fwrite(adtsData, 1, adtsDataSize, adtsFile);

            fwrite(buffer, 1, frameInfo.bytesconsumed, adtsFile);
        }

        if (buffer) free(buffer);

        if (!noGapless)
        {
            if (sampleId == 0) dur = 0;

            if (useAacLength || (timescale != samplerate)) {
                sample_count = frameInfo.samples;
            } else {
                sample_count = (unsigned int)(dur * frameInfo.channels);

                if (!useAacLength && !initial && (sampleId < numSamples/2) && (sample_count != frameInfo.samples))
                {
                    fprintf(stderr, "MP4 seems to have incorrect frame duration, using values from AAC data.\n");
                    useAacLength = 1;
                    sample_count = frameInfo.samples;
                }
            }

            if (initial && (sample_count < framesize*frameInfo.channels) && (frameInfo.samples > sample_count))
                delay = frameInfo.samples - sample_count;
        } else {
            sample_count = frameInfo.samples;
        }

        /* open the sound file now that the number of channels are known */
        if (first_time && !frameInfo.error)
        {
            /* print some channel info */
/*            print_channel_info(&frameInfo); */

            if (!adts_out)
            {
                /* open output file */
                if(!to_stdout)
                {
                    aufile = open_audio_file(sndfile, frameInfo.samplerate, frameInfo.channels,
                        outputFormat, fileType, aacChannelConfig2wavexChannelMask(&frameInfo));
                } else {
                    aufile = open_audio_file("-", frameInfo.samplerate, frameInfo.channels,
                        outputFormat, fileType, aacChannelConfig2wavexChannelMask(&frameInfo));
                }
                if (aufile == NULL)
                {
                    faacDecClose(hDecoder);
                    mp4ff_close(infile);
                    free(mp4cb);
                    fclose(mp4File);
                    return 0;
                }
            }
            first_time = 0;
        }

        if (sample_count > 0) initial = 0;

        percent = min((int)(sampleId*100)/numSamples, 100);
        if (percent > old_percent)
        {
            old_percent = percent;
            sprintf(percents, "%d%% decoding %s.", percent, mp4file);
            fprintf(stderr, "%s\r", percents);
        }
        
        /**
         * TODO: We need to be able to audio_play intervald. So, find out how we
         * can skip to X time in file and audio_play for Y seconds.
         */
        printf ("sampleId: %li\tnumSamples: %li\tsampple_count_pr_channel: %d\n", sampleId, numSamples, sample_count/channels);

		/**
		 * Callback function to audio.c. This is where we handle the
		 * messages that tells this thread what to do.
		 */
		callback (data, dur);	/* struct_audio_data_t, duration in milliseconds */

        if ((frameInfo.error == 0) && (sample_count > 0) && (!adts_out))
        {
			/**
			 * libao_play(sample_buffer, sample_count); gives us wrong output.
			 * Sounds like each sample is played twice as fast as it ought to be.
			 * If we multiply with 2, output is OK.
			 * Possible solution: We'll have to multiply with the number of channels.
			 */
			libao_play(data->device, sample_buffer, sample_count*channels);
        }

        if (frameInfo.error > 0)
        {
            fprintf(stderr, "Warning: %s\n",
                faacDecGetErrorMessage(frameInfo.error));
        }
    }

    faacDecClose(hDecoder);

    if (adts_out == 1)
    {
        fclose(adtsFile);
    }


	/* Closing daisylibao */
	libao_terminate(data->device);

    mp4ff_close(infile);

    if (!first_time && !adts_out)
        close_audio_file(aufile);

    free(mp4cb);
    fclose(mp4File);

    return frameInfo.error;
}



/* ************************************************	*
 * 			Global function declarations  		  	*
 * ************************************************	*/

/**
 * Play a file.
 * @param struct_audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 */
bool aac_play (struct_audio_data_t *data)
{
    int result;
    int infoOnly = 0;
    int writeToStdio = 0;
    int object_type = LC;
    int def_srate = 0;
    int downMatrix = 0;
    int format = 1;
    int outputFormat = FAAD_FMT_16BIT;
    int adts_out = 0;
    int old_format = 0;
    int mp4file = 0;
    int noGapless = 0;
    char aacFileName[STRLEN];
    char audioFileName[STRLEN];
    char adtsFileName[STRLEN];
    unsigned char header[8];
    float length = 0;
    FILE *hMP4File;

    strncpy (aacFileName, data->file_name, STRLEN);

    /* check for mp4 file */
    mp4file = 0;
    hMP4File = fopen(aacFileName, "rb");
    if (!hMP4File)
    {
        fprintf(stderr, "Error opening file: %s\n", aacFileName);
        return 1;
    }
    fread(header, 1, 8, hMP4File);
    fclose(hMP4File);
    if (header[4] == 'f' && header[5] == 't' && header[6] == 'y' && header[7] == 'p')
        mp4file = 1;


	/**
	 * Based on what sort of file this is, we call with decodeMP4file or decodeAACfile
	 */
    if (mp4file)
    {
		if (DEBUG_AAC) report ("File is mp4", REP_DEBUG);
		result = decodeMP4file(data, aacFileName, audioFileName, adtsFileName, writeToStdio,
            outputFormat, format, downMatrix, noGapless, infoOnly, adts_out, &length);
    } else {
		if (DEBUG_AAC) report ("File is aac", REP_DEBUG);
		result = decodeAACfile(data, aacFileName, audioFileName, adtsFileName, writeToStdio,
            def_srate, object_type, outputFormat, format, downMatrix, infoOnly, adts_out,
            old_format, &length);
    }
	return 0;
}

/**
 * Initiates AAC decoder.
 * @param struct_audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 */
void aac_initiate (struct_audio_data_t *data)
{
	data = data;
	printf ("aac_init: \n");
	if (DEBUG_AAC) report ("aac_init", REP_DEBUG);
}

/**
 * Terminates AAC decoder.
 * @param struct_audio_data_t _data - the struct_audio_data_t that is passed to almost all functions in the audio module.
 */
void aac_terminate (struct_audio_data_t *data)
{
	data = data;
	printf ("aac_close: \n");
	if (DEBUG_AAC) report ("aac_close", REP_DEBUG);
}



#endif	/* WITH_AAC */
